%
%
%  K-means Spectral Clustering Function for Imaris 
%
%  Tristan McRae 2018
%  References - http://www.scs2.net/next/files/courses/iic/ImarisOpenLaunchExtract.pdf
%
%  Installation:
%
%  - Copy this file into the XTensions folder in the Imaris installation directory
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%        <Submenu name='Custom Extensions'>
%          <Item name='Spectral Unmixing' icon='Matlab' tooltip='Split overlapping fluorophores to each have their own channel'>
%            <Command>MatlabXT::XTSpectralUnmixing(%i)</Command>
%          </Item>
%        </Submenu>
%      </Menu>
%    </CustomTools>
%  
%
%  Description:
%
%   This XTension separates pixels into clusters based on their spectral
%   signatures
%   This helps to discriminate between pixels that share a dominant
%   channel but belong to different parts of the cell
%
%


function XTSpectralUnmixing(aImarisApplicationID)
global num_channels num_timesteps x_dim y_dim z_dim unmixed_image  mixed_image %infer_channel_names median_filter extra_channel_names
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%Load data
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ID
disp('Loading data into MATLAB (for large datasets, this may take some time)')

disp('Setting up connection')
% Set up connection	between	Imaris and MATLAB
if	isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
    vImarisApplication = aImarisApplicationID;
else
    % connect to Imaris interface
    javaaddpath ImarisLib.jar
    vImarisLib = ImarisLib;
    if	ischar(aImarisApplicationID)
        aImarisApplicationID = round(str2double(aImarisApplicationID));
    end
    vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
end

%Set up progress bar
vProgressDisplay = waitbar(0,'LUMoS:Loading Data...');

disp('Finding the dataset properties')
% Get the vDataSet object
vDataSet = vImarisApplication.GetDataSet();
% Get the data type
switch char(vDataSet.GetType())
    case 'eTypeUInt8', datatype = 'uint8';
    case 'eTypeUInt16', datatype = 'uint16';
    case 'eTypeFloat', datatype = 'single';
    otherwise, error('Bad value for vDataSet.GetType()');
end
% Allocate memory
disp('Allocating Memory')
x_dim = vDataSet.GetSizeX();
y_dim = vDataSet.GetSizeY();
z_dim = vDataSet.GetSizeZ();
set_x(x_dim);
set_y(y_dim);
set_z(z_dim);
num_channels = vDataSet.GetSizeC();
set_num_channels(num_channels)
num_timesteps = vDataSet.GetSizeT();
set_t(num_timesteps)
mixed_image = zeros([x_dim, y_dim, z_dim, num_channels, num_timesteps], datatype);
basic_stack = zeros([x_dim, y_dim, z_dim], datatype);
% Get the pixel data
for t = 1:num_timesteps
    waitbar(0.1*t/num_timesteps);
    disp(strcat('Reading timestep_', int2str(t), ' of_', int2str(num_timesteps)))
    pause(0.01) %let the display catch up
    for c = 1:num_channels
        switch char(vDataSet.GetType())
            case 'eTypeUInt8'
                arr = vDataSet.GetDataVolumeAs1DArrayBytes(c-1, t-1); 
                basic_stack(:) = typecast(arr, 'uint8');
                mixed_image(:, :, :, c, t) = basic_stack;
                
            case 'eTypeUInt16'
                arr	= vDataSet.GetDataVolumeAs1DArrayShorts(c-1, t-1);
                basic_stack(:) = typecast(arr, 'uint16');
                mixed_image(:, :, :, c, t) = basic_stack;
                
            case 'eTypeFloat'
                basic_stack(:) = vDataSet.GetDataVolumeAs1DArrayFloats(c-1, t-1);
                mixed_image(:, :, :, c, t) = basic_stack; 
            otherwise, error('Bad value for type');
        end
    end
end
set_mixed_image(mixed_image)
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

%for c = 1:num_channels
%    channel_names(c) = vDataSet.GetChannelName(c-1);
%end



spectral_unmixing_gui();%[unmixed_image, infer_channel_names, extra_channel_names]  = spectral_unmixing_gui()%(num_channels, num_timesteps, x_dim, y_dim, z_dim, stack);
disp('returning image')
close all;
vProgressDisplay = waitbar(0.8,'Returning Image to Imaris');
unmixed_image = get_unmixed_image;
%infer_channel_names = get_infer;
global median_filter infer_channel_names extra_channel_names

return_results_to_imaris(unmixed_image, vDataSet, num_timesteps, vImarisApplication, infer_channel_names, median_filter, extra_channel_names);

close all

end

function set_num_channels(num_c)
    global num_channels
    num_channels = num_c;
end
function num_c = get_num_channels()
    global num_channels
    num_c = num_channels; 
end

function set_mixed_image(mi)
    global mixed_image
    mixed_image = mi; 
end
function mi = get_mixed_image()
    global mixed_image
    mi = mixed_image;
end
function set_unmixed_image(ui)
    global unmixed_image
    unmixed_image = ui; 
end
function ui = get_unmixed_image()
    global unmixed_image
    ui = unmixed_image;
end
function set_x(x)
    global x_dim
    x_dim = x;
end
function x = get_x()
    global x_dim
    x = x_dim;
end
function set_y(y)
    global y_dim
    y_dim = y;
end
function y = get_y()
    global y_dim
    y = y_dim;
end
function set_z(z)
    global z_dim
    z_dim = z;
end
function z = get_z()
    global z_dim
    z = z_dim;
end
function set_t(t)
    global num_timesteps
    num_timesteps = t;
end
function t = get_t()
    global num_timesteps
    t = num_timesteps;
end
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%Prompt user for unmixing details
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [] = spectral_unmixing_gui()%num_channels, num_timesteps, x_dim, y_dim, z_dim, mixed_image)
    global background_removal_only vals 
    close all;

    h.f = figure('units','pixels','position',[700,500,200,100],...
                 'toolbar','none','menu','none', 'Name', 'LUMoS Task', 'NumberTitle', 'off');

    h.t = uicontrol('style', 'text','units', 'pixels',...
        'position',[0, 60, 150, 30], 'string', 'Select a task');           

    h.d = uicontrol('style', 'popupmenu', 'units', 'pixels',...
        'position',[10, 30, 150, 30], 'string', 'Spectral Unmixing|Background Removal Only');
        %'position',[10, 30, 90, 30], 'string', 'K-means|SIMI (coming soon!)|NNMF (coming soon!)');

    h.p = uicontrol('style','pushbutton','units','pixels',...
                    'position',[40,5,70,20],'string','OK',...
                    'callback',@d_call);
    uiwait(h.f)
    % Dropdown callback
    function d_call(varargin)
        vals = get(h.d,'Value');
        switch vals
            case 1
                background_removal_only = false;
            case 2
                background_removal_only = true;
        end
        close all;
    end

    switch vals
        case 1
            num_channels = get_num_channels();
            k_means_gui(num_channels);
        case 2
            initialize_background_removal;
            k_means_unmixing();
    end


end

function[] = initialize_background_removal()
    global replicates scaled_signatures multi_fluorophore_pixels infer_channel_names median_filter representitive_timestep ...
        representitive_slice maximum_iterations channels_to_unmix num_fluorophores_to_unmix num_channels...
        num_timesteps z_dim

    channels_to_unmix = 1:num_channels
    num_fluorophores_to_unmix = num_channels
    replicates = 10
    scaled_signatures = true
    multi_fluorophore_pixels = false
    infer_channel_names = true
    median_filter = true
    representitive_timestep = floor(num_timesteps/2)+1
    representitive_slice = floor(z_dim/2)+1
    maximum_iterations  = 100

end

function [] = k_means_gui(num_channels)%k_means_gui(num_channels, num_timesteps, x_dim, y_dim, z_dim, mixed_image)
    global h  

    height = 190;
    width = 300; %max(20+40*num_channels, 300)
    h.f = figure('units','pixels','position',[700,500,width,height],...
                 'toolbar','none','menu','none', 'Name', 'LUMoS Settings', 'NumberTitle', 'off');

    h.t(1) = uicontrol('style', 'text','units', 'pixels',...
        'position',[-60, height-40, 400, 30], 'string', 'How many fluorophores do you want to unmix?');

    h.num_fluorophores_to_unmix = uicontrol('style', 'edit','units', 'pixels',...
        'position',[50, height-60, 40, 30]);        

    h.t(2) = uicontrol('style', 'text','units', 'pixels',...
        'position',[0, height-100, 200, 30], 'string', 'Select which channels to unmix');

    for i = 1:num_channels
        h.c(i) = uicontrol('style','checkbox','units','pixels',...
                        'position',[40+40*(i-1),height-110,50,15],'string',int2str(i));
    end
    
                    
    pb = uicontrol('Style','togglebutton','position',[20,height-150,100,20],...
                        'Callback',@togglebutton1_Callback, 'string', 'Advanced Options');
                    
                    
    h.p = uicontrol('style','pushbutton','units','pixels',...
                    'position',[40,5,70,20],'string','OK',...
                    'callback',@p_call, 'UserData', struct('num_channels',num_channels));
    uiwait(h.f)


end

function [] = k_means_unmixing()
    global maximum_iterations num_channels representitive_slice channels_to_unmix replicates scaled_signatures num_fluorophores_to_unmix multi_fluorophore_pixels...
        infer_channel_names median_filter representitive_timestep... 
        unmixed_image extra_channel_names background_removal_only num_coloc_channels...

    x_dim = get_x;
    y_dim = get_y;
    z_dim = get_z;
    num_timesteps = get_t;
    
    mixed_image = get_mixed_image;
    use_input_intensities = true; %Setting this to false is good for debugging where certain clusters are
    channels = 1:num_channels
    channels_to_leave = channels(~ismember(channels, channels_to_unmix));
    num_fluorophores_to_unmix = num_fluorophores_to_unmix+1; %add background cluster automatically
    m=2; %https://en.wikipedia.org/wiki/Fuzzy_clustering#Fuzzy_C-means_clustering
    channels_to_unmix
    
    disp(num_timesteps)
    disp(size(mixed_image))
    
    %It doesn't like doing kmeans on unsigned integers (understandibly)
    mixed_image = double(mixed_image);
    %save original pixel intensities for later
    unscaled_mixed_image = mixed_image;
      
    
    if median_filter
       disp('Applying median filter')
       for c = 1:num_channels
          for t = 1:num_timesteps
            mixed_image(:, :, :, c, t) = medfilt3(mixed_image(:, :, :, c, t), [3, 3, 3]);
          end
       end
    end
    
    %Select only representitive timestep and slice
    if num_timesteps>1
        reordered_mixed_image = permute(mixed_image, [4 1 2 3 5]);
        reordered_mixed_image = reordered_mixed_image(:, :, :, :, representitive_timestep);
    else
        reordered_mixed_image = permute(mixed_image, [4 1 2 3]);
    end
    disp(size(reordered_mixed_image))
    
    if z_dim>1
        reordered_mixed_image = reordered_mixed_image(:, :, :, representitive_slice);
    end
    disp(size(reordered_mixed_image))
    

    
    if scaled_signatures && length(channels_to_unmix)>1
        %change so that it only includes channels to unmix
        reordered_mixed_image_channels_to_unmix = reordered_mixed_image(channels_to_unmix, :, :);
        mixed_image_channels_to_unmix = mixed_image(:, :, :, channels_to_unmix, :);
        disp('Scaling')
        for c = 1:num_channels
            reordered_mixed_image(c, :, :) = reordered_mixed_image(c, :, :)./sum(reordered_mixed_image_channels_to_unmix, 1);
            mixed_image(:, :, :, c, :) = mixed_image(:, :, :, c, :)./sum(mixed_image_channels_to_unmix, 4); 
        end
        reordered_mixed_image(isnan(reordered_mixed_image)) = 0;
        mixed_image(isnan(mixed_image)) = 0;
    end
    
    %Normalize pixel intensities
    for i = 1:length(channels_to_unmix) %scale even the channel you're leaving alone so that softmaxing works. may not be needed now
        c = channels_to_unmix(i);
        mixed_image_channel = reordered_mixed_image(c, :, :);
        mu = mean(mixed_image_channel(:))
        sigma = std(mixed_image_channel(:))
        reordered_mixed_image(c, :, :) = (reordered_mixed_image(c, :, :)-mu)/sigma;
        mixed_image(:, :, :, c, :) = (mixed_image(:, :, :, c, :)-mu)/sigma; %Get z-scores for entire image
    end 

    disp('Finished scaling intensities')
    
    pixel_array = reordered_mixed_image(:, :);
    pixel_array = permute(pixel_array, [2 1]);
    
    disp('reordered')

    %select just the channels you want to unmix
    pixel_array = pixel_array(:, channels_to_unmix); 

    close all;    
    vProgressDisplay = waitbar(0.2,'LUMoS:Running K-Means...');
    %This really does take a long time for big datasets. Running on just
    %one z-slice could help that
    disp('Running K-Means. The display may freeze but it is still working behind the scenes. This may take a few minutes for large datasets.')
    if multi_fluorophore_pixels
        min_objFunc = inf
        for i=1:replicates
            [cluster_centroids_i, U, objFunc] = fcm(pixel_array, num_fluorophores_to_unmix, [m, maximum_iterations, 1e-3, true]);
            if objFunc(end) <= min_objFunc
                min_objFunc = objFunc(end)
                cluster_centroids = cluster_centroids_i
            end
        end
    else
        [cluster_indices, cluster_centroids, sumd, D] = kmeans(pixel_array, num_fluorophores_to_unmix, 'Replicates', replicates, 'MaxIter', maximum_iterations, 'Display', 'iter');%, 'Start', initial_centroids);
    end
    disp('Finished Running k-means')
    close all;
    vProgressDisplay = waitbar(0.5,'LUMoS:Finished Running K-Means...');
    pause(0.01) %let the display catch up
    
    
    cluster_centroids_copy = cluster_centroids
    
    if infer_channel_names && ~background_removal_only
        extra_channel_names = {'initializer'}
        for c = 1:length(channels_to_unmix)
            [~, closest_cluster] = max(cluster_centroids_copy(:, c));
            cluster_centroids(c, :) = cluster_centroids_copy(closest_cluster, :)
            cluster_centroids_copy(closest_cluster, :) = [];
        end
        
        for c = length(channels_to_unmix)+1:num_fluorophores_to_unmix
            [~, next_cluster] = max(cluster_centroids_copy(:, 1));
            cluster_centroids(c, :) = cluster_centroids_copy(next_cluster, :)
            cluster_centroids_copy(next_cluster, :) = [];
            
            if c < length(channels_to_unmix)+1+num_coloc_channels
                extra_channel_names{c-length(channels_to_unmix)} = 'colocalization'
            else
                extra_channel_names{c-length(channels_to_unmix)} = 'background'
            end
            
        end
        
    else
        extra_channel_names = {}
    end

    cluster_centroids
    
    cluster_totals = zeros(num_fluorophores_to_unmix, 1);

    if not(multi_fluorophore_pixels)
        for i = 1:num_fluorophores_to_unmix
            cluster_totals(i) = sum(cluster_indices==i);
        end
        cluster_totals
    end

    
    %creating image
    distances = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix);
    
    cluster_center_broadcast = zeros(x_dim, y_dim, z_dim, length(channels_to_unmix), num_fluorophores_to_unmix);
    for cluster = 1:num_fluorophores_to_unmix
        cluster_center = reshape(cluster_centroids(cluster, :), 1, 1, 1, length(channels_to_unmix));
        cluster_center_broadcast(:, :, :, :, cluster) = repmat(cluster_center, x_dim, y_dim, z_dim);
    end
        
    if background_removal_only
        num_output_channels = num_fluorophores_to_unmix-1;
        unmixed_image = zeros(x_dim, y_dim, z_dim, num_output_channels, num_timesteps);
        %cluster_weights_image = zeros(x_dim, y_dim, z_dim, num_channels, num_timesteps);
        
        
        [minimum_cluster_mean, background_cluster_index] = min(max(cluster_centroids, [], 2));
        background_cluster = cluster_centroids(background_cluster_index, :)
        
        unmixed_image(:, :, :, :, :) = unscaled_mixed_image;
        
        cluster_weights = zeros(x_dim*y_dim*z_dim*num_timesteps, num_fluorophores_to_unmix);
        for t = 1:num_timesteps
            cluster_weights_timestep = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix); %so that you don't get weird tails effects
            disp(strcat('applying clustering to timestep ', int2str(t), ' of ', int2str(num_timesteps)))
            waitbar(0.5+0.3*t/num_timesteps);
            for cluster = 1:num_fluorophores_to_unmix   
                pixel_values = double(mixed_image(:, :, :, channels_to_unmix, t));
                diff = pixel_values - cluster_center_broadcast(:, :, :, :, cluster);
                intermediate_step = vecnorm(diff, 2, 4); 
                distances(:, :, :, cluster) = squeeze(intermediate_step);
            end
            pause(0.01)
            [~, cluster_indices] = min(distances, [], length(size(distances)));
            for c = 1:num_output_channels
                unmixed_image(:, :, :, c, t) = unmixed_image(:, :, :, c, t) .* (cluster_indices~=background_cluster_index);
            end
        end
        
        
        
    else
        cluster_weights_image = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix, num_timesteps);
        num_output_channels = num_fluorophores_to_unmix+length(channels_to_leave);
        
        cluster_weights_timestep = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix); %was originally unm_channels instead of num_fluorophores to unmix
        
        disp('Broadcasting cluster centers')
        %This takes a while as well when there are a lot of pixels. Is there a
        %way to do with matrix math?

        
        close all;
        vProgressDisplay = waitbar(0.5,'Applying clustering to data');


        if multi_fluorophore_pixels
            %cluster_weights = zeros(x_dim*y_dim*z_dim*num_timesteps, num_fluorophores_to_unmix);
            for t = 1:num_timesteps
                cluster_weights_timestep = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix);
                disp(strcat('applying clustering to timestep ', int2str(t), ' of ', int2str(num_timesteps)))
                waitbar(0.5+0.3*t/num_timesteps);
                for cluster = 1:num_fluorophores_to_unmix   
                    pixel_values = double(mixed_image(:, :, :, channels_to_unmix, t));
                    diff = pixel_values - cluster_center_broadcast(:, :, :, :, cluster);
                    intermediate_step = vecnorm(diff, 2, 4); 
                    distances(:, :, :, cluster) = squeeze(intermediate_step);
                end
                pause(0.01)
                
                inverse_square_weights = 1./distances.^(2/(m-1));
                inverse_square_weights_pixel_sums = sum(inverse_square_weights, 4);
                
                for c = 1:num_fluorophores_to_unmix  
                    cluster_weights_timestep(:, :, :, c) = inverse_square_weights(:, :, :, c)./inverse_square_weights_pixel_sums;
                end
                
                %cluster_weights_timestep(:) = cluster_weights(:);
                cluster_weights_image(:, :, :, :, t) = cluster_weights_timestep; 
            end
        else
            cluster_weights = zeros(x_dim*y_dim*z_dim*num_timesteps, num_fluorophores_to_unmix);
            for t = 1:num_timesteps
                cluster_weights_timestep = zeros(x_dim, y_dim, z_dim, num_fluorophores_to_unmix); %so that you don't get weird tails effects
                disp(strcat('applying clustering to timestep ', int2str(t), ' of ', int2str(num_timesteps)))
                waitbar(0.5+0.3*t/num_timesteps);
                for cluster = 1:num_fluorophores_to_unmix   
                    pixel_values = double(mixed_image(:, :, :, channels_to_unmix, t));
                    diff = pixel_values - cluster_center_broadcast(:, :, :, :, cluster);
                    intermediate_step = vecnorm(diff, 2, 4); 
                    distances(:, :, :, cluster) = squeeze(intermediate_step);
                end
                pause(0.01)
                [~, cluster_indices] = min(distances, [], length(size(distances)));

                for x=1:x_dim
                    for y = 1:y_dim
                        for z = 1:z_dim
                            cluster_weights_timestep(x,y,z,cluster_indices(x,y,z)) = 1;
                        end
                    end
                end
                cluster_weights_image(:, :, :, :, t) = cluster_weights_timestep;
            end
        end

        %cluster_weights for each pixel sum to the total intensity of the channels
        %being unmixed after this step
        if use_input_intensities
            if multi_fluorophore_pixels   
                for c = 1:num_fluorophores_to_unmix
                    cluster_weights_image(:, :, :, c, :) = cluster_weights_image(:, :, :, c, :) .*sum(unscaled_mixed_image(:, :, :, channels_to_unmix, :),  4); 
                end
            else
                for c = 1:num_fluorophores_to_unmix
                    cluster_weights_image(:, :, :, c, :) = cluster_weights_image(:, :, :, c, :) .*max(unscaled_mixed_image(:, :, :, channels_to_unmix, :), [],  4); 
                end
            end
        else
            cluster_weights_image = cluster_weights_image * double(max(unscaled_mixed_image(:)));
        end

        unmixed_image = zeros(x_dim, y_dim, z_dim, num_output_channels, num_timesteps);

        if length(channels_to_leave)>0
            unmixed_image(:, :, :, 1:length(channels_to_leave), :) = unscaled_mixed_image(:, :, :, channels_to_leave, :);
        end
        unmixed_image(:, :, :, length(channels_to_leave)+1:end, :) = cluster_weights_image;
    
    end
    set_unmixed_image(unmixed_image)

end %breakpoint

function return_results_to_imaris(unmixed_image, vDataSet, num_timesteps, vImarisApplication, infer_channel_names, median_filter, extra_channel_names)

% Create a new channel
dims = size(unmixed_image);
if length(dims) <= 2
	num_output_channels = 1; 
else
	num_output_channels = dims(4);
end
vDataSet.SetSizeC(num_output_channels);  


for t = 1:num_timesteps
    disp(strcat('returning timestep ', int2str(t), ' of ', int2str(num_timesteps)))
    waitbar(0.8+0.2*t/num_timesteps)
    pause(0.01)
    for c = 1:num_output_channels
        %get channel name
        basic_stack = unmixed_image(:, :, :, c, t);
        switch char(vDataSet.GetType())
            case 'eTypeUInt8'
                vDataSet.SetDataVolumeAs1DArrayBytes(uint8(basic_stack(:)), c-1, t-1); 
            case 'eTypeUInt16'
                vDataSet.SetDataVolumeAs1DArrayShorts(uint16(basic_stack(:)), c-1, t-1);
            case 'eTypeFloat'
                vDataSet.SetDataVolumeAs1DArrayFloats(single(basic_stack(:)), c-1, t-1);
            otherwise, error('Bad value for type');
        end
        %rename channel
        if infer_channel_names 
            if (c>(num_output_channels-length(extra_channel_names)))
                vDataSet.SetChannelName(c-1, extra_channel_names{c-(num_output_channels-length(extra_channel_names))});
            end
        else
            channel_name = strcat('Ch',int2str(c));
            vDataSet.SetChannelName(c-1, channel_name);
        end   
    end
end

close all;
vProgressDisplay = waitbar(1,'Done');
close(vProgressDisplay) %End progress bar display
vImarisApplication.SetDataSet(vDataSet);

end

function togglebutton1_Callback(hObject, eventdata)
	global h advanced_options %num_channels
    button_state = get(hObject,'Value');
    if button_state == get(hObject,'Max')
        close all; 
        advanced_options = true;
        num_channels = get_num_channels
        
        height = 530;
        width = 500; %max(20+40*num_channels, 300)
        h.f = figure('units','pixels','position',[700,400,width,height],... 
                     'toolbar','none','menu','none', 'Name', 'LUMoS Settings', 'NumberTitle', 'off');

        h.t(1) = uicontrol('style', 'text','units', 'pixels',...
            'position',[-61, height-40, 400, 30], 'string', 'How many fluorophores do you want to unmix?');      

        h.num_fluorophores_to_unmix = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-60, 40, 30]);
        
        h.t(2) = uicontrol('style', 'text','units', 'pixels',...
        'position',[-19, height-100, 400, 30], 'string', 'How many instances of colocalization do you want to separate?');

        h.num_coloc = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-120, 40, 30]);
    
        h.t(3) = uicontrol('style', 'text','units', 'pixels',...
            'position',[1, height-160, 200, 30], 'string', 'Select which channels to unmix');

        for i = 1:num_channels
            h.c(i) = uicontrol('style','checkbox','units','pixels',...
                            'position',[40+40*(i-1),height-170,50,15],'string',int2str(i));
        end

        h.t(4) = uicontrol('style', 'text','units', 'pixels',...
            'position',[-3, height-210, 400, 30], 'string', 'Number of replicates (runs k-means multiple times and uses best result)');  

        h.replicates = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-230, 40, 30]); 

        h.t(5) = uicontrol('style', 'text','units', 'pixels',...
            'position',[-70, height-270, 300, 30], 'string', 'Representitive timestep');      

        h.representitive_timestep = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-290, 40, 30]); 

        h.t(6) = uicontrol('style', 'text','units', 'pixels',...
            'position',[-30, height-330, 200, 30], 'string', 'Representitive slice'); 

        h.representitive_slice = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-350, 40, 30]); 

        h.t(7) = uicontrol('style', 'text','units', 'pixels',...
            'position',[-5, height-390, 200, 30], 'string', 'Maximum number of iterations'); 

        h.max_iter = uicontrol('style', 'edit','units', 'pixels',...
            'position',[50, height-410, 40, 30]); 

        h.scale = uicontrol('style','checkbox','units','pixels',...
                            'position',[20,height-440,400,15],'string','Unmix based on pixel intensity ratio (instead of absolute channel values)');

        h.multi_fluorophore = uicontrol('style','checkbox','units','pixels',...
                            'position',[20,height-460,400,15],'string','Allow multiple fluorophores to occupy a single pixel');

        h.infer_channel_names = uicontrol('style','checkbox','units','pixels',...
                            'position',[20,height-480,400,15],'string','Infer channel names');
                        
        h.median_filter = uicontrol('style','checkbox','units','pixels',...
                            'position',[20,height-500,400,15],'string','Apply median filter');

        h.p = uicontrol('style','pushbutton','units','pixels',...
                        'position',[40,5,70,20],'string','OK',...
                        'callback',@p_call);
                    
        uiwait(h.f)


    elseif button_state == get(hObject,'Min')
        close(figure(3))
    end
end

function p_call(varargin)
    global h maximum_iterations channels_to_unmix replicates scaled_signatures num_fluorophores_to_unmix multi_fluorophore_pixels...
        infer_channel_names median_filter representitive_timestep representitive_slice advanced_options num_timesteps z_dim num_coloc_channels
    %global h num_channels num_timesteps x_dim y_dim z_dim mixed_image
    vals = get(h.c,'Value');
    if length(vals)>1
        channels_to_unmix = find([vals{:}]);
    else
        channels_to_unmix = find(vals);
    end
    num_coloc_channels =str2num(get(h.num_coloc, 'String'));
    num_fluorophores_to_unmix = str2num(get(h.num_fluorophores_to_unmix, 'String'))+num_coloc_channels;
    
    if advanced_options
        replicates = str2num(get(h.replicates, 'String'))
        scaled_signatures = get(h.scale,'Value')
        multi_fluorophore_pixels = get(h.multi_fluorophore,'Value')
        infer_channel_names = get(h.infer_channel_names,'Value')
        median_filter = get(h.median_filter,'Value')
        representitive_timestep = str2num(get(h.representitive_timestep, 'String'))
        representitive_slice = str2num(get(h.representitive_slice, 'String'))
        maximum_iterations  = str2num(get(h.max_iter, 'String'))
    else
        replicates = 10
        scaled_signatures = true
        multi_fluorophore_pixels = false
        infer_channel_names = true
        median_filter = true
        representitive_timestep = floor(num_timesteps/2)+1
        representitive_slice = floor(z_dim/2)+1
        maximum_iterations  = 100
    end


    k_means_unmixing();%[unmixed_image, extra_channel_names] = k_means_unmixing()%num_channels, num_timesteps, x_dim, y_dim, z_dim, num_fluorophores_to_unmix, channels_to_unmix, mixed_image, replicates,...
        %scaled_signatures, multi_fluorophore_pixels, representitive_timestep, representitive_slice, infer_channel_names, max_iter);
    uiresume
end

function B = medfilt3(A, filterSize)
    %   Copyright 2016-2017 The MathWorks, Inc.

    padopt = 'symmetric';
    
    if isempty(A)
        B = A;
        return
    end

    padMap = containers.Map( ...
        {'zeros', 'indexed', 'replicate', 'symmetric'}, ...
        {0, 1, 2, 3});

    radius = (filterSize - 1) / 2;

    if ismatrix(A)
        B = medfilt2(A, filterSize(1:2), padopt);
    else
        B = ordfilt3D(A, 14, padopt);
    end
end

function n = vecnorm(A,p,dim)
    n = sqrt(sum(A.^p, dim));
end

function B = medianfilter3d(A, radius)


    [xmax,ymax,zmax] = size(A);
    B=A;
    for x = 1:xmax
        for y = 1:ymax
            for z=1:zmax
                neighbors = get_neighbors(A, radius, x, y, z);
                B(x,y,z) = median(neighbors);
            end
        end
    end
end

function [Vr] = ordfilt3D(V0,ord,padoption)
    % ordfilt3D:    Perform 3-D order-statistic filtering on 26 neighbors
    %
    %   [Vr] = ordfilt3D(V0,ord,padoption)
    %          use 26 neighbors
    %       ord = 14 <=> median filtering
    %       ord = 1 <=> min
    %       ord = [1 27] <=> [min max]
    %       padoption: same as in padarray
    %
    % Olivier Salvado, Case Western Reserve University, 16Aug04



    if ~exist('padoption','var')
        padoption = 'replicate';
    end


    %%
    % special care for uint8
    if isa(V0,'uint8')
        V = uint8(padarray(V0,[1 1 1],padoption));
        S = size(V);
        Vn = uint8(zeros(S(1),S(2),S(3),26));  % all the neighbor
    else
        V = single(padarray(V0,[1 1 1],padoption));
        S = size(V);
        Vn = single(zeros(S(1),S(2),S(3),26));  % all the neighbor
    end

    %%
    % build the neighboord
    Vn(:,:,:,1) = V;
    i = 1:S(1); ip1 = [i(2:end) i(end)]; im1 = [i(1) i(1:end-1)];
    j = 1:S(2); jp1 = [j(2:end) j(end)]; jm1 = [j(1) j(1:end-1)];
    k = 1:S(3); kp1 = [k(2:end) k(end)]; km1 = [k(1) k(1:end-1)];

    %%
    % left
    Vn(:,:,:,2)     = V(im1    ,jm1    ,km1);
    Vn(:,:,:,3)     = V(im1    ,j      ,km1);
    Vn(:,:,:,4)     = V(im1    ,jp1    ,km1);

    Vn(:,:,:,5)     = V(im1    ,jm1    ,k);
    Vn(:,:,:,6)     = V(im1    ,j      ,k);
    Vn(:,:,:,7)     = V(im1    ,jp1    ,k);

    Vn(:,:,:,8)     = V(im1    ,jm1    ,kp1);
    Vn(:,:,:,9)     = V(im1    ,j      ,kp1);
    Vn(:,:,:,10)    = V(im1    ,jp1    ,kp1);

    %%
    % right
    Vn(:,:,:,11)    = V(ip1    ,jm1    ,km1);
    Vn(:,:,:,12)    = V(ip1    ,j      ,km1);
    Vn(:,:,:,13)    = V(ip1    ,jp1    ,km1);

    Vn(:,:,:,14)    = V(ip1    ,jm1    ,k);
    Vn(:,:,:,15)    = V(ip1    ,j      ,k);
    Vn(:,:,:,16)    = V(ip1    ,jp1    ,k);

    Vn(:,:,:,17)    = V(ip1    ,jm1    ,kp1);
    Vn(:,:,:,18)    = V(ip1    ,j      ,kp1);
    Vn(:,:,:,19)    = V(ip1    ,jp1    ,kp1);

    %%
    % top
    Vn(:,:,:,20)    = V(i       ,jm1    ,kp1);
    Vn(:,:,:,21)    = V(i       ,j      ,kp1);
    Vn(:,:,:,22)    = V(i       ,jp1    ,kp1);

    %%
    % bottom
    Vn(:,:,:,23)    = V(i       ,jm1    ,km1);
    Vn(:,:,:,24)    = V(i       ,j      ,km1);
    Vn(:,:,:,25)    = V(i       ,jp1    ,km1);

    %%
    % front
    Vn(:,:,:,26)    = V(i       ,jp1    ,k);

    %%
    % back
    Vn(:,:,:,27)    = V(i       ,jm1    ,k);

    %%
    % perform the processing
    Vn = sort(Vn,4);
    Vr = Vn(:,:,:,ord);


    %%
    % remove padding on the 3 first dimensions
    Vr = Vr(2:end-1,2:end-1,2:end-1,:);

end

